#ifndef _CAMERA_H_
#define _CAMERA_H_

/*------------------------------------------------------------------------------
 Forward declaration
 -----------------------------------------------------------------------------*/
class cCamera;
	
/*------------------------------------------------------------------------------
 Includes to complete declaration of this class
 -----------------------------------------------------------------------------*/
#include <GL/gl.h>
#include <GL/glu.h>
#include "cfrustum.h"


#define FRUSTUM_STACK_LENGTH 8


/**
 * cCamera - This is our camera class. It provides support for frustum culling. Also it 
 * has support for stereo view. Stereo view we are using is called passive stereo. 
 * This camera class also supports to kind of farPlane definition. You can define farPlane
 * and also viewing distance. By calling setUp() the projection matrix of frsutum with
 * far plane at viewing distance will be calculated. This will be used for frustum culling.
 * farPlane - property defines the standard far plane after that OpenGL do his clipping.
 * If you didn't understand for what you can use it, just define this to properties with same values.
 *
 * This class also supports frustum stack. That means you can push current frustum on the stack
 * and work with another one. After your work is done (e.g. offscreen rendering) you restore the old
 * frustum from the stack. The storing and restoring produce no extra costs.
 *
 * @author: Art Tevs (arti_tevs@yahoo.de)
 */
class cCamera {
private:
	
	// Frustum
	cFrustum frustum[FRUSTUM_STACK_LENGTH];
	int frustumStackPos;
	
	// Properties of our viewing frustum
	int viewPortWidth;
	int viewPortHeight;
	float nearPlane;
	float fieldOfView;
	float farPlane;
	float viewingDistance;
		
	// current matrices
	mat4 modelview;
	mat4 projection;
	mat4 viewingDistProjection;
	
	// Camera orientation
	vec3 m_vPosition;
	vec3 m_vView;
	vec3 m_vUpVector;
	
	// camera to world orientation (usefull for billboards)	
	vec3 m_DirectionVector;
	vec3 m_RightVector;
	vec3 m_UpVector;
		
	// Stereo-View
	bool	useStereo;
	bool	currentView;				// Current view Left = 0, Right = 1;
	float	halfEyeSeparation;
	mat4	asymmetricProjectionLeft;		// Frustum for Left Eye  (Projection Matrix)
	mat4	asymmetricProjectionRight;		// 		   for Right Eye (Projection Matrix)
	bool	widthHalfed;
	float	eyeSeparation, focalLength;
 
public:
	
	// Our camera constructor
	cCamera();
	virtual ~cCamera();
	
	// Stereo View
	const mat4&	getLeftEyeProjection(){return asymmetricProjectionLeft;}
	const mat4&	getRightEyeProjection(){return asymmetricProjectionRight;}
	void		calculateEyeProjections(float eyeSep, float focLength);
	void		calculateEyeProjections(){calculateEyeProjections(eyeSeparation, focalLength);}
	void		setUseStereo(bool b);		// set if we want to use stereo mode
	bool		isUsingStereoView(){return useStereo;}
	float		getEyeSeparation(){return eyeSeparation;}
	float		getFocalLength() {return focalLength;}
	bool		isUsingRightEye(){return currentView;}	// if true, so we have right eye view active
	void		activateLeftEye(bool setNewVieport = true);
	void		activateRightEye(bool setNewVieport = true);
	void		deactivateEyeViews(bool setNewVieport = true);
	
	// Get matrices
	const mat4& getProjection(){return projection;}
	const mat4& getModelview(){return modelview;}
	
	// Get frustum
	cFrustum*	getFrustum(){return &frustum[frustumStackPos];}
	void		pushFrustum();
	void		calculateFrustum();
	void		popFrustum();
	void		calculateDirections();
	
	// get/set other camera properties
	void 	setViewingDistance (float d);
	float 	getViewingDistance (){return viewingDistance;}
	int		getWidth(){return viewPortWidth;}
	int		getHeight(){return viewPortHeight;}
	void	setFOV (float f){fieldOfView = f;if(useStereo)calculateEyeProjections(eyeSeparation,focalLength);}
	float	getFOV(){return fieldOfView;}
	void	setWidth(int w){viewPortWidth = w;}
	void	setHeight(int h){viewPortHeight = h;}
	
	// These are are data access functions for our camera's private data
	vec3 Position() {   return m_vPosition;     }
	vec3 View()     {   return m_vView;         }
	vec3 UpVector() {   return m_vUpVector;     }
	
	// get basis of our camera to world matrix
	vec3 ViewDir()  {   return m_DirectionVector;     }
	vec3 RightDir() {   return m_RightVector;         }
	vec3 UpDir()    {   return m_UpVector;     }
	
	void setUp();
	
	// This changes the position, view, and up vector of the camera.
	// This is primarily used for initialization
	void PositionCamera(float positionX, float positionY, float positionZ,
						float viewX,     float viewY,     float viewZ,
						float upVectorX, float upVectorY, float upVectorZ);
	
	void PositionCamera(const vec3& pos,const vec3& view,const vec3& up){
			PositionCamera(pos.x, pos.y, pos.z, view.x,view.y,view.z, up.x,up.y,up.z);}
	
	// This will position camera
	void PositionCameraByDirection (float posX, float posY, float posZ,
			float dirX, float dirY, float dirZ,
			float upVectorX, float upVectorY, float upVectorZ);
	
	// Move camera to new position
	void MoveTo (const vec3& pos){PositionCamera(pos,m_vView,m_vUpVector);}	
	void MoveTo (float posX, float posY, float posZ){
			PositionCamera(vec3(posX,posY,posZ),m_vView,m_vUpVector);} 
	
	// Set new orientation
	void ViewAt (const vec3& view){PositionCamera(m_vPosition, view, m_vUpVector);}
	void ViewAt (float x,float y,float z){PositionCamera(m_vPosition,vec3(x,y,z), m_vUpVector);}
	
	// roll camera
	void ChangeUp (const vec3& up){PositionCamera(m_vPosition, m_vView, up);}
	void ChangeUp (float x,float y,float z){PositionCamera(m_vPosition,m_vView,vec3(x,y,z));}
	
	// This rotates the camera's view around the position depending on the values passed in.
	void RotateView(float angle, float X, float Y, float Z);
	
	// This moves the camera's view by the mouse movements (First person view)
	void SetViewByMouse();
	
	// This rotates the camera around a point (I.E. your character).
	void RotateAroundPoint(vec3 vCenter, float angle, float X, float Y, float Z);
	
	// This strafes the camera left or right depending on the speed (+/-)
	void StrafeCamera(float speed);
	
	// This will move the camera forward or backward depending on the speed
	void MoveCamera(float speed);
	
	// This updates the camera's view and other data (Should be called each frame)
	void Update();
	
	// This uses gluLookAt() to tell OpenGL where to look
	void Look();
		
	// Neccessary stuff to get Z-Value of Depth-Buffer in certain point
	float calcZValue(const vec3& point);
	
};

#endif
